SystemOrganization addCategory: #CenteredIntervalTree!

Object subclass: #CenteredIntervalTree
	instanceVariableNames: 'iMidLeft iMidRight leftSubTree rightSubTree xMid'
	classVariableNames: 'EmptyCollection'
	poolDictionaries: ''
	category: 'CenteredIntervalTree'!

!CenteredIntervalTree commentStamp: 'PaulDeBruicker 10/14/2011 09:00' prior: 0!
This is faster to find all intersecting timespans from a collection of timespans than selecting all intersections across a collection. It is also faster than a DateTree.  The querying method #intervalsThatInclude: returns a collection, which may need to be converted into a set.  

If you use the Timespan specifc creation method (#fromTimespans:) the timespans are converted into Intervals.  The creation and querying of trees created from Intervals is much faster than that of Timespans.  The querying methods (#intervalsThatIncludeTime:,  #intervalsThatIncludeTimespan:) return a set.  



The implementation and ideas behind the structure come from here: https://secure.wikimedia.org/wikipedia/en/wiki/Interval_tree

and: 

http://brettwalenz.org/2010/09/28/centered-interval-trees/!

!CenteredIntervalTree classSide methodsFor: 'instance creation' stamp: 'PaulDeBruicker 10/12/2011 06:55'!
from: aCollectionOfIntervals

	^ aCollectionOfIntervals isEmpty
		ifTrue: [ CenteredIntervalTreeSentinelNode default ]
		ifFalse: [ self new createFrom: aCollectionOfIntervals ]! !

!CenteredIntervalTree classSide methodsFor: 'instance creation' stamp: 'PaulDeBruicker 10/13/2011 13:20'!
fromTimespans: aCollectionOfTimespans


	 ^self new createFrom: (aCollectionOfTimespans collect:[:each| each start asSeconds to: each end asSeconds])! !

!CenteredIntervalTree classSide methodsFor: 'as yet unclassified' stamp: 'PaulDeBruicker 10/14/2011 08:56'!
initialize 
	EmptyCollection:=OrderedCollection new.! !

!CenteredIntervalTree methodsFor: 'initialization' stamp: 'PaulDeBruicker 10/13/2011 10:54'!
checkForOverlap: aRange
	| col |
	col := OrderedCollection new.

	iMidLeft
		do: [ :each | (aRange includes:each)
				ifTrue: [ col add: each ]
				ifFalse: [ ^ col ]] .
			^col
! !

!CenteredIntervalTree methodsFor: 'initialization' stamp: 'PaulDeBruicker 10/14/2011 08:56'!
checkLeftSideFor: aPoint
	^ aPoint < xMid
		ifTrue: [ (self selectFrom: iMidLeft a: aPoint) , (leftSubTree intervalsThatInclude: aPoint) ]
		ifFalse: [ EmptyCollection ]! !

!CenteredIntervalTree methodsFor: 'initialization' stamp: 'PaulDeBruicker 10/14/2011 08:57'!
checkLeftSideForRange: aRange
	^ aRange start < xMid
		ifTrue: [ (self selectFrom: iMidLeft a: aRange end) , (leftSubTree intervalsThatIncludeRange: aRange) ]
		ifFalse: [ EmptyCollection  ]! !

!CenteredIntervalTree methodsFor: 'initialization' stamp: 'PaulDeBruicker 10/14/2011 08:57'!
checkRightSideFor: aPoint
	^ aPoint > xMid
		ifTrue: [ (self selectFrom: iMidRight a: aPoint)  , (rightSubTree intervalsThatInclude: aPoint) ]
		ifFalse: [ EmptyCollection  ]! !

!CenteredIntervalTree methodsFor: 'initialization' stamp: 'PaulDeBruicker 10/14/2011 08:57'!
checkRightSideForRange: aRange
	^ aRange end > xMid
		ifTrue: [ (self selectFrom: iMidRight a: aRange start) , (rightSubTree intervalsThatIncludeRange: aRange) ]
		ifFalse: [ EmptyCollection  ]! !

!CenteredIntervalTree methodsFor: 'initialization' stamp: 'PaulDeBruicker 10/12/2011 06:50'!
contents
	^ iMidLeft ,  leftSubTree contents , rightSubTree contents! !

!CenteredIntervalTree methodsFor: 'initialization' stamp: 'PaulDeBruicker 10/13/2011 19:14'!
createFrom: aCollectionOfIntervals 
 |iMid iLeft iRight|

	xMid:=(aCollectionOfIntervals collect:[:each | each end]) median.
		iMidLeft:=(aCollectionOfIntervals select:[:each | each start<= xMid and:[each end >= xMid]]) asSortedCollection:[:a :c | a start < c start].
		iMidRight:=iMidLeft asSortedCollection:[:a :c | a end < c end].
		leftSubTree:=CenteredIntervalTree from:(aCollectionOfIntervals select: [:each | each end < xMid]).
		rightSubTree:=CenteredIntervalTree from:(aCollectionOfIntervals select: [:each | each start > xMid]).
	! !

!CenteredIntervalTree methodsFor: 'initialization' stamp: 'PaulDeBruicker 10/11/2011 15:59'!
end
	^self contents last end! !

!CenteredIntervalTree methodsFor: 'initialization' stamp: 'PaulDeBruicker 10/11/2011 10:51'!
initialize 
	iMidLeft:=OrderedCollection new.
	iMidRight:=OrderedCollection new.! !

!CenteredIntervalTree methodsFor: 'querying' stamp: 'PaulDeBruicker 10/11/2011 12:21'!
intervalsThatInclude: aPointOrRange
	^ (aPointOrRange respondsTo: #start)
		ifTrue: [ self intervalsThatIncludeRange: aPointOrRange ]
		ifFalse: [ self intervalsThatIncludePoint: aPointOrRange ]! !

!CenteredIntervalTree methodsFor: 'initialization' stamp: 'PaulDeBruicker 10/12/2011 06:52'!
intervalsThatIncludePoint: aPoint
	^ (self checkLeftSideFor: aPoint) , (self checkRightSideFor: aPoint)! !

!CenteredIntervalTree methodsFor: 'initialization' stamp: 'PaulDeBruicker 10/13/2011 09:10'!
intervalsThatIncludeRange: aRange

	^
		(self checkLeftSideForRange:aRange) , (self checkRightSideForRange: aRange) , (self checkForOverlap: aRange)
		
! !

!CenteredIntervalTree methodsFor: 'querying' stamp: 'PaulDeBruicker 10/13/2011 13:49'!
intervalsThatIncludeTime: aTime
	
	 ^(self intervalsThatIncludePoint: aTime asSeconds) collect:[:each | Timespan fromSecondsInterval: each]! !

!CenteredIntervalTree methodsFor: 'querying' stamp: 'PaulDeBruicker 10/13/2011 13:54'!
intervalsThatIncludeTimespan: aTimespan
	^ (self intervalsThatIncludeRange: (aTimespan start asSeconds to: aTimespan end asSeconds))
		collect: [ :each | Timespan fromSecondsInterval: each]! !

!CenteredIntervalTree methodsFor: 'initialization' stamp: 'PaulDeBruicker 10/11/2011 17:12'!
isEmpty
	^iMidLeft isEmpty and:[iMidRight isEmpty]! !

!CenteredIntervalTree methodsFor: 'initialization' stamp: 'PaulDeBruicker 10/12/2011 06:48'!
isLeaf
	^ false! !

!CenteredIntervalTree methodsFor: 'initialization' stamp: 'PaulDeBruicker 10/11/2011 16:48'!
matches: anIntervalTree
	|matches|
	matches:=OrderedCollection new.
	anIntervalTree contents do:[:each |matches addAll: (self intervalsThatIncludeRange: each) ].
	^matches! !

!CenteredIntervalTree methodsFor: 'initialization' stamp: 'PaulDeBruicker 10/11/2011 17:38'!
medianValue
	^xMid! !

!CenteredIntervalTree methodsFor: 'initialization' stamp: 'PaulDeBruicker 10/11/2011 16:31'!
notEmpty
	^self isEmpty not! !

!CenteredIntervalTree methodsFor: 'initialization' stamp: 'PaulDeBruicker 10/13/2011 09:40'!
selectFrom: aSortedCollection a: pointValue
^(aSortedCollection select:[:each | each includes: pointValue ]) asOrderedCollection ! !

!CenteredIntervalTree methodsFor: 'initialization' stamp: 'PaulDeBruicker 10/12/2011 06:49'!
size
	^ self contents size! !

!CenteredIntervalTree methodsFor: 'initialization' stamp: 'PaulDeBruicker 10/11/2011 15:58'!
start
	^self contents first start! !

Object subclass: #CenteredIntervalTreeSentinelNode
	instanceVariableNames: 'emptyCollection'
	classVariableNames: 'Default'
	poolDictionaries: ''
	category: 'CenteredIntervalTree'!

!CenteredIntervalTreeSentinelNode commentStamp: 'PaulDeBruicker 10/14/2011 08:54' prior: 0!
This is just an empty node that responds to the same methods as CenteredIntervalTree but terminates the recursive process of walking the tree branches.  !

!CenteredIntervalTreeSentinelNode classSide methodsFor: 'as yet unclassified' stamp: 'PaulDeBruicker 10/12/2011 06:54'!
default
	^Default ifNil:[Default:=self new]! !

!CenteredIntervalTreeSentinelNode methodsFor: 'as yet unclassified' stamp: 'PaulDeBruicker 10/12/2011 06:50'!
contents
	^emptyCollection ! !

!CenteredIntervalTreeSentinelNode methodsFor: 'as yet unclassified' stamp: 'PaulDeBruicker 10/12/2011 06:49'!
iMidLeft
	^emptyCollection ! !

!CenteredIntervalTreeSentinelNode methodsFor: 'as yet unclassified' stamp: 'PaulDeBruicker 10/12/2011 06:49'!
iMidRight
	^emptyCollection ! !

!CenteredIntervalTreeSentinelNode methodsFor: 'as yet unclassified' stamp: 'PaulDeBruicker 10/12/2011 06:49'!
initialize 
	emptyCollection :=OrderedCollection new.! !

!CenteredIntervalTreeSentinelNode methodsFor: 'testing' stamp: 'PaulDeBruicker 10/12/2011 06:56'!
intervalsThatInclude: aTimespan 
	^emptyCollection ! !

!CenteredIntervalTreeSentinelNode methodsFor: 'testing' stamp: 'PaulDeBruicker 10/13/2011 19:09'!
intervalsThatIncludePoint: aTimespan 
	^emptyCollection ! !

!CenteredIntervalTreeSentinelNode methodsFor: 'testing' stamp: 'PaulDeBruicker 10/13/2011 19:09'!
intervalsThatIncludeRange: aTimespan 
	^emptyCollection ! !

!CenteredIntervalTreeSentinelNode methodsFor: 'as yet unclassified' stamp: 'PaulDeBruicker 10/12/2011 06:51'!
isEmpty 
	^true! !

!CenteredIntervalTreeSentinelNode methodsFor: 'as yet unclassified' stamp: 'PaulDeBruicker 10/12/2011 06:47'!
isLeaf
	^true! !

!CenteredIntervalTreeSentinelNode methodsFor: 'as yet unclassified' stamp: 'PaulDeBruicker 10/12/2011 06:51'!
notEmpty 
	^false! !

!CenteredIntervalTreeSentinelNode methodsFor: 'as yet unclassified' stamp: 'PaulDeBruicker 10/12/2011 06:48'!
size
	^0! !

!Timespan class methodsFor: '*centeredintervaltree' stamp: 'PaulDeBruicker 10/13/2011 13:55'!
fromSecondsInterval: anInterval
	^ self
		starting: (DateAndTime fromSeconds: anInterval start)
		ending: (DateAndTime fromSeconds: (anInterval end + 1))! !

TestCase subclass: #CenteredIntervalTreeTest
	instanceVariableNames: 'timespans pointInTime timespan intervalDatabase'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CenteredIntervalTree'!

!CenteredIntervalTreeTest methodsFor: 'as yet unclassified' stamp: 'PaulDeBruicker 10/13/2011 10:51'!
setUp

pointInTime:=DateAndTime year:2011 month:10 day: 13 hour: 12 minute: 0 second: 0 offset: Duration zero.
timespans:=(1 to: 100) collect:[:each |(Timespan starting: (pointInTime + (each-50) days) duration: (8 hours))] .

timespan:=Timespan starting: pointInTime + 15 minutes duration: 10 hours.

intervalDatabase:=(1to:10) collect:[:each | each to:(each * 20)].
! !

!CenteredIntervalTreeTest methodsFor: 'as yet unclassified' stamp: 'PaulDeBruicker 10/13/2011 13:46'!
testBasicQuery
	|cit int |
	cit:=CenteredIntervalTree fromTimespans: timespans.
	self assert: (cit intervalsThatIncludeTimespan: timespan) notEmpty.
	self assert: (cit intervalsThatIncludeTime: pointInTime) notEmpty.
	cit:=CenteredIntervalTree from: intervalDatabase.
	int:=intervalDatabase atRandom.
	self assert: (cit intervalsThatInclude: int) notEmpty.
	self assert: (cit intervalsThatInclude: int start) notEmpty.
	self assert: (cit intervalsThatInclude: int end) notEmpty.! !

!CenteredIntervalTreeTest methodsFor: 'as yet unclassified' stamp: 'PaulDeBruicker 10/13/2011 13:46'!
testCreateFrom
	|cit |
	cit:=CenteredIntervalTree fromTimespans: timespans.
	self assert: cit medianValue notNil.
	self assert: cit isEmpty equals: false.
! !

!CenteredIntervalTreeTest methodsFor: 'as yet unclassified' stamp: 'PaulDeBruicker 10/13/2011 13:45'!
testEnclosedTimespan
	|cit tspan |
	cit:=CenteredIntervalTree fromTimespans: timespans.
	tspan:=Timespan starting: pointInTime + 1 hour  duration: 1 hour.
	self assert: (cit intervalsThatIncludeTimespan: tspan) notEmpty. ! !

!CenteredIntervalTreeTest methodsFor: 'as yet unclassified' stamp: 'PaulDeBruicker 10/13/2011 13:47'!
testFaillingQuery
	|cit tspan |
	cit:=CenteredIntervalTree fromTimespans: timespans.
	tspan:=Timespan starting: pointInTime + 5000 days duration: 1 hour.
	self assert: (cit intervalsThatIncludeTimespan: tspan) isEmpty. "tspan is after the latest timespan in the tree"
	self assert: (cit intervalsThatIncludeTime: pointInTime + 5000 days) isEmpty.
	tspan:=Timespan starting: pointInTime - 5000 days duration: 1 hour.	
	self assert: (cit intervalsThatIncludeTimespan: tspan) isEmpty. "tspan is before  the earliest timespan in the tree"
	self assert: (cit intervalsThatIncludeTime: pointInTime - 5000 days) isEmpty.
	! !

!CenteredIntervalTreeTest methodsFor: 'as yet unclassified' stamp: 'PaulDeBruicker 10/13/2011 13:48'!
testFindAllTimespans
	| cit col |
	cit:=CenteredIntervalTree fromTimespans: timespans.
	col:=OrderedCollection new.
	timespans do:[:each | col addAll: (cit intervalsThatIncludeTimespan: each) asSet].
	self assert: col size equals: timespans size.
	self assert: col sort first equals: timespans sort first
! !

!CenteredIntervalTreeTest methodsFor: 'as yet unclassified' stamp: 'PaulDeBruicker 10/13/2011 13:56'!
testQuerySpeed
	| cit intersectionTime citTime |
	intersectionTime := [ 10 timesRepeat: [ timespans select: [ :each | (each intersection: timespan) notNil ] ] ]
		timeToRun.
	cit := CenteredIntervalTree fromTimespans: timespans.
	citTime := [ cit := CenteredIntervalTree fromTimespans: timespans.
		10 timesRepeat: [ cit intervalsThatIncludeTimespan: timespan ] ] timeToRun.
	self assert: citTime < intersectionTime! !

!CenteredIntervalTreeTest methodsFor: 'as yet unclassified' stamp: 'PaulDeBruicker 10/13/2011 13:57'!
testStartAndEndMatched
	|cit tspan |
	cit:=CenteredIntervalTree fromTimespans: timespans.
	tspan:=Timespan starting: pointInTime duration: 1 hour.
	self assert: (cit intervalsThatIncludeTimespan: tspan) notEmpty. "start matches"
		tspan:=Timespan starting: pointInTime duration: 8 hour.
	self assert: (cit intervalsThatIncludeTimespan: tspan) notEmpty. "exact overlap" 
		tspan:=Timespan starting: pointInTime + 7 hours duration: 1 hour.
	self assert: (cit intervalsThatIncludeTimespan: tspan) notEmpty. "end matches"
	! !

!Interval methodsFor: '*centeredintervaltree' stamp: 'PaulDeBruicker 10/13/2011 09:02'!
end
	^self last! !

!Interval methodsFor: '*centeredintervaltree' stamp: 'PaulDeBruicker 10/13/2011 10:57'!
includes: aNumberOrInterval
	^(aNumberOrInterval respondsTo: #end)
		ifTrue: [ start <= aNumberOrInterval end and: [ stop >= aNumberOrInterval start ] ]
		ifFalse: [ self rangeIncludes: aNumberOrInterval ]! !

!Interval methodsFor: '*centeredintervaltree' stamp: 'PaulDeBruicker 10/13/2011 09:03'!
start
	^self first! !
CenteredIntervalTree initialize!
